#![no_std]
#![feature(int_roundings)]
#![feature(lang_items)]
#![feature(abi_x86_interrupt)]
#![allow(internal_features)]

use core::arch::asm;
use lazy_static::lazy_static;
use rand::{rngs::SmallRng, SeedableRng};
use spin::Mutex;

use drivers::{pic::PICDriver, vga::Vga13h};
use idt::{InterruptDescriptorTable, InterruptStackFrame};
use os::{
	thunk::ThunkData,
	util::{sti, without_interrupts},
};
use tetrisboard::{FallingTetromino, TetrisBoard};

mod idt;
mod os;
mod tetrisboard;

#[macro_use]
mod drivers;

const PIC_OFFSET: u8 = 32;

pub(crate) static VGA: Mutex<Vga13h> = Mutex::new(unsafe { Vga13h::new() });
pub(crate) static PIC: Mutex<PICDriver> = Mutex::new(PICDriver::new(PIC_OFFSET, PIC_OFFSET + 8));
pub(crate) static TICK_COUNTER: Mutex<u32> = Mutex::new(0);
pub(crate) static BOARD: Mutex<TetrisBoard> = Mutex::new(TetrisBoard::new());
pub(crate) static FALLING: Mutex<Option<FallingTetromino>> = Mutex::new(None);
pub(crate) static LAST_INPUT: Mutex<Option<InputKey>> = Mutex::new(None);

lazy_static! {
	static ref RNG: Mutex<SmallRng> = Mutex::new(SmallRng::seed_from_u64(1337));
	static ref IDT: InterruptDescriptorTable = {
		let mut idt = InterruptDescriptorTable::new();

		idt.divide_error.set_handler_fn(divide_handler);
		idt.double_fault.set_handler_fn(double_fault_handler);
		idt.interrupts[InterruptIndex::Timer.as_idx()].set_handler_fn(timer_handler);
		idt.interrupts[InterruptIndex::Keyboard.as_idx()].set_handler_fn(keyboard_handler);

		idt
	};
}

#[derive(Debug, Clone, Copy)]
#[repr(u8)]
pub enum InterruptIndex {
	Timer = PIC_OFFSET,
	Keyboard,
}

impl InterruptIndex {
	fn as_u8(self) -> u8 {
		self as u8
	}

	fn as_idx(self) -> usize {
		usize::from(self.as_u8() - PIC_OFFSET)
	}
}

extern "x86-interrupt" fn divide_handler(stack_frame: InterruptStackFrame) {
	println!("DIVIDE ERROR {:?}", stack_frame);
}

unsafe fn inb(port: u32) -> u8 {
	let mut out;

	asm!(
		"in al, dx",
		out("al") out,
		in("dx") port,
	);

	return out;
}

#[derive(Debug, Clone, Copy, PartialEq, Eq)]
enum InputKey {
	Left,
	Right,
	Up,
	Down,
}

extern "x86-interrupt" fn keyboard_handler(_stack_frame: InterruptStackFrame) {
	let scancode = unsafe { inb(0x60) };
	println!("{:x?}", scancode);

	let key = match scancode {
		0x11 => Some(InputKey::Up),
		0x1E => Some(InputKey::Left),
		0x1F => Some(InputKey::Down),
		0x20 => Some(InputKey::Right),
		0xE0 => {
			let scancode = unsafe { inb(0x60) };
			println!("e {:x?}", scancode);
			None
		}
		_ => None,
	};

	*LAST_INPUT.lock() = key;
	PIC.lock().send_eoi(InterruptIndex::Keyboard.as_u8());
}

extern "x86-interrupt" fn timer_handler(_stack_frame: InterruptStackFrame) {
	let mut t = TICK_COUNTER.lock();
	*t = (*t).wrapping_add(1);
	PIC.lock().send_eoi(InterruptIndex::Timer.as_u8());
}

extern "x86-interrupt" fn double_fault_handler(
	stack_frame: InterruptStackFrame,
	error_code: u32,
) -> ! {
	panic!("DOUBLE FAULT (err = 0x{error_code:x}\n{:#?}", stack_frame);
}

#[no_mangle]
pub unsafe extern "C" fn start(
	_boot_disk: usize,
	thunk10: extern "C" fn(),
	_thunk13: extern "C" fn(),
	_thunk15: extern "C" fn(),
	_thunk16: extern "C" fn(),
) -> ! {
	println!("Entered Rust, serial ready.");

	{
		// Set vga mode
		let mut data = ThunkData::new();
		data.eax = 0x13;
		data.with(thunk10);
	}

	{
		// Disable cursor
		let mut data = ThunkData::new();
		data.eax = 0x0100;
		data.ecx = 0x3F00;
		data.with(thunk10);
	}

	{
		// Initialize IDT
		IDT.load();
		let mut pic = PIC.lock();
		pic.init();
	}

	// We're ready for interrupts,
	// enable them
	sti();

	let mut last_t = 0;
	loop {
		// All locks use `without_interrupts`
		// to prevent deadlocks
		let t = without_interrupts(|| {
			let l = TICK_COUNTER.lock();
			let t = *l;
			drop(l);
			return t;
		});

		if t == last_t {
			continue;
		}
		last_t = t;

		// Handle user input
		without_interrupts(|| {
			if let Some(fall) = &mut *FALLING.lock() {
				let board = BOARD.lock();
				let mut fall_test = fall.clone();
				let mut last_input = LAST_INPUT.lock();

				match *last_input {
					Some(InputKey::Up) => {
						fall_test.rotate_cw();
						if board.tetromino_valid(&fall_test) {
							fall.rotate_cw()
						};
					}

					Some(InputKey::Down) => {
						fall_test.rotate_ccw();
						if board.tetromino_valid(&fall_test) {
							fall.rotate_ccw()
						};
					}

					Some(InputKey::Left) => {
						fall_test.translate(-1, 0);
						if board.tetromino_valid(&fall_test) {
							fall.translate(-1, 0);
						};
					}

					Some(InputKey::Right) => {
						fall_test.translate(1, 0);
						if board.tetromino_valid(&fall_test) {
							fall.translate(1, 0);
						};
					}

					_ => {}
				}

				// Clear last input, it was handled.
				*last_input = None;
			}
		});

		// Update board
		without_interrupts(|| {
			let mut v = VGA.lock();
			let mut board = BOARD.lock();
			let mut fall = FALLING.lock();

			v.swap();
			board.draw(&mut v, fall.as_ref());

			if let Some(fall_inner) = fall.as_mut() {
				if t % 5 == 0 {
					let mut fall_test = fall_inner.clone();
					fall_test.translate(0, 1);

					if board.tetromino_valid(&fall_test) {
						fall_inner.translate(0, 1);
					} else {
						let mut x = None;
						core::mem::swap(&mut x, &mut fall);
						board.place_tetromino(x.unwrap());
						board.collapse();
					}
				}
			} else {
				*fall = Some(FallingTetromino::random(5, 1))
			}
		})
	}
}
